@@ -12,6 +12,8 @@
     :license: MIT, see LICENSE for more details
 """
 
+from typing import Optional, Set, Callable
+
 import functools
 import logging
 import socket as Socket
@@ -19,10 +21,12 @@ import ssl
 import weakref
 import uuid
 
-import xml.etree.ElementTree
+from asyncio import iscoroutinefunction, wait
+
+import xml.etree.ElementTree as ET
 
 from slixmpp.xmlstream.asyncio import asyncio
-from slixmpp.xmlstream import tostring, highlight
+from slixmpp.xmlstream import tostring
 from slixmpp.xmlstream.stanzabase import StanzaBase, ElementBase
 from slixmpp.xmlstream.resolver import resolve, default_resolver
 
@@ -30,6 +34,10 @@ from slixmpp.xmlstream.resolver import resolve, default_resolver
 RESPONSE_TIMEOUT = 30
 
 log = logging.getLogger(__name__)
+class ContinueQueue(Exception):
+    """
+    Exception raised in the send queue to "continue" from within an inner loop
+    """
 
 class NotConnectedError(Exception):
     """
@@ -64,12 +72,12 @@ class XMLStream(asyncio.BaseProtocol):
     :param int port: The port to use for the connection. Defaults to 0.
     """
 
-    def __init__(self, socket=None, host='', port=0):
+    def __init__(self, host='', port=0):
         # The asyncio.Transport object provided by the connection_made()
         # callback when we are connected
         self.transport = None
 
-        # The socket the is used internally by the transport object
+        # The socket that is used internally by the transport object
         self.socket = None
 
         self.connect_loop_wait = 0
@@ -81,6 +89,8 @@ class XMLStream(asyncio.BaseProtocol):
         self.force_starttls = None
         self.disable_starttls = None
 
+        self.waiting_queue = asyncio.Queue()
+
         # A dict of {name: handle}
         self.scheduled_events = {}
 
@@ -199,10 +209,8 @@ class XMLStream(asyncio.BaseProtocol):
         self.__event_handlers = {}
         self.__filters = {'in': [], 'out': [], 'out_sync': []}
 
-        self._id = 0
-
-        #: We use an ID prefix to ensure that all ID values are unique.
-        self._id_prefix = '%s-' % uuid.uuid4()
+        # Current connection attempt (Future)
+        self._current_connection_attempt = None
 
         #: A list of DNS results that have not yet been tried.
         self.dns_answers = None
@@ -212,6 +220,9 @@ class XMLStream(asyncio.BaseProtocol):
         #: ``_xmpp-client._tcp`` service.
         self.dns_service = None
 
+        #: The reason why we are disconnecting from the server
+        self.disconnect_reason = None
+
         #: An asyncio Future being done when the stream is disconnected.
         self.disconnected = asyncio.Future()
 
@@ -235,12 +246,7 @@ class XMLStream(asyncio.BaseProtocol):
         ID values. Using this method ensures that all new ID values
         are unique in this stream.
         """
-        self._id += 1
-        return self.get_id()
-
-    def get_id(self):
-        """Return the current unique stream ID in hexadecimal form."""
-        return "%s%X" % (self._id_prefix, self._id)
+        return uuid.uuid4().hex
 
     def connect(self, host='', port=0, use_ssl=False,
                 force_starttls=True, disable_starttls=False):
@@ -265,6 +271,13 @@ class XMLStream(asyncio.BaseProtocol):
                                localhost
 
         """
+        asyncio.ensure_future(
+            self.run_filters(),
+            loop=self.loop,
+        )
+        self.disconnect_reason = None
+        self.cancel_connection_attempt()
+        self.connect_loop_wait = 0
         if host and port:
             self.address = (host, int(port))
         try:
@@ -281,15 +294,22 @@ class XMLStream(asyncio.BaseProtocol):
             self.disable_starttls = disable_starttls
 
         self.event("connecting")
-        asyncio.async(self._connect_routine())
+        self._current_connection_attempt = asyncio.ensure_future(
+            self._connect_routine(),
+            loop=self.loop,
+        )
 
-    @asyncio.coroutine
-    def _connect_routine(self):
+    async def _connect_routine(self):
         self.event_when_connected = "connected"
 
-        record = yield from self.pick_dns_answer(self.default_domain)
+        if self.connect_loop_wait > 0:
+            self.event('reconnect_delay', self.connect_loop_wait)
+            await asyncio.sleep(self.connect_loop_wait, loop=self.loop)
+
+        record = await self.pick_dns_answer(self.default_domain)
         if record is not None:
-            host, address, port = record
+            host, address, dns_port = record
+            port = dns_port if dns_port else self.address[1]
             self.address = (address, port)
             self._service_name = host
         else:
@@ -297,23 +317,33 @@ class XMLStream(asyncio.BaseProtocol):
             # and try (host, port) as a last resort
             self.dns_answers = None
 
-        yield from asyncio.sleep(self.connect_loop_wait)
+        if self.use_ssl:
+            ssl_context = self.get_ssl_context()
+        else:
+            ssl_context = None
+
+        if self._current_connection_attempt is None:
+            return
         try:
-            yield from self.loop.create_connection(lambda: self,
+            await self.loop.create_connection(lambda: self,
                                                    self.address[0],
                                                    self.address[1],
-                                                   ssl=self.use_ssl,
+                                                   ssl=ssl_context,
                                                    server_hostname=self.default_domain if self.use_ssl else None)
+            self.connect_loop_wait = 0
         except Socket.gaierror as e:
             self.event('connection_failed',
                        'No DNS record available for %s' % self.default_domain)
         except OSError as e:
             log.debug('Connection failed: %s', e)
             self.event("connection_failed", e)
+            if self._current_connection_attempt is None:
+                return
             self.connect_loop_wait = self.connect_loop_wait * 2 + 1
-            asyncio.async(self._connect_routine())
-        else:
-            self.connect_loop_wait = 0
+            self._current_connection_attempt = asyncio.ensure_future(
+                self._connect_routine(),
+                loop=self.loop,
+            )
 
     def process(self, *, forever=True, timeout=None):
         """Process all the available XMPP events (receiving or sending data on the
@@ -328,10 +358,10 @@ class XMLStream(asyncio.BaseProtocol):
             else:
                 self.loop.run_until_complete(self.disconnected)
         else:
-            tasks = [asyncio.sleep(timeout)]
+            tasks = [asyncio.sleep(timeout, loop=self.loop)]
             if not forever:
                 tasks.append(self.disconnected)
-            self.loop.run_until_complete(asyncio.wait(tasks))
+            self.loop.run_until_complete(asyncio.wait(tasks, loop=self.loop))
 
     def init_parser(self):
         """init the XML parser. The parser must always be reset for each new
@@ -339,14 +369,18 @@ class XMLStream(asyncio.BaseProtocol):
         """
         self.xml_depth = 0
         self.xml_root = None
-        self.parser = xml.etree.ElementTree.XMLPullParser(("start", "end"))
+        self.parser = ET.XMLPullParser(("start", "end"))
 
     def connection_made(self, transport):
         """Called when the TCP connection has been established with the server
         """
         self.event(self.event_when_connected)
         self.transport = transport
-        self.socket = self.transport.get_extra_info("socket")
+        self.socket = self.transport.get_extra_info(
+            "ssl_object",
+            default=self.transport.get_extra_info("socket")
+        )
+        self._current_connection_attempt = None
         self.init_parser()
         self.send_raw(self.stream_header)
         self.dns_answers = None
@@ -358,33 +392,55 @@ class XMLStream(asyncio.BaseProtocol):
         event.  This could trigger one or more event (a stanza is received,
         the stream is opened, etc).
         """
+        if self.parser is None:
+            log.warning('Received data before the connection is established: %r',
+                        data)
+            return
         self.parser.feed(data)
-        for event, xml in self.parser.read_events():
-            if event == 'start':
-                if self.xml_depth == 0:
-                    # We have received the start of the root element.
-                    self.xml_root = xml
-                    log.debug('[33;1mRECV[0m: %s', highlight(tostring(self.xml_root, xmlns=self.default_ns,
-                                                         stream=self,
-                                                         top_level=True,
-                                                         open_only=True)))
-                    self.start_stream_handler(self.xml_root)
-                self.xml_depth += 1
-            if event == 'end':
-                self.xml_depth -= 1
-                if self.xml_depth == 0:
-                    # The stream's root element has closed,
-                    # terminating the stream.
-                    log.debug("End of stream received")
-                    self.abort()
-                elif self.xml_depth == 1:
-                    # A stanza is an XML element that is a direct child of
-                    # the root element, hence the check of depth == 1
-                    self._spawn_event(xml)
-                    if self.xml_root is not None:
-                        # Keep the root element empty of children to
-                        # save on memory use.
-                        self.xml_root.clear()
+        try:
+            for event, xml in self.parser.read_events():
+                if event == 'start':
+                    if self.xml_depth == 0:
+                        # We have received the start of the root element.
+                        self.xml_root = xml
+                        log.debug('RECV: %s', tostring(self.xml_root,
+                                                       xmlns=self.default_ns,
+                                                       stream=self,
+                                                       top_level=True,
+                                                       open_only=True))
+                        self.start_stream_handler(self.xml_root)
+                    self.xml_depth += 1
+                if event == 'end':
+                    self.xml_depth -= 1
+                    if self.xml_depth == 0:
+                        # The stream's root element has closed,
+                        # terminating the stream.
+                        self.end_session_on_disconnect = True
+                        log.debug("End of stream received")
+                        self.disconnect_reason = "End of stream"
+                        self.abort()
+                    elif self.xml_depth == 1:
+                        # A stanza is an XML element that is a direct child of
+                        # the root element, hence the check of depth == 1
+                        self._spawn_event(xml)
+                        if self.xml_root is not None:
+                            # Keep the root element empty of children to
+                            # save on memory use.
+                            self.xml_root.clear()
+        except ET.ParseError:
+            log.error('Parse error: %r', data)
+
+            # Due to cyclic dependencies, this can’t be imported at the module
+            # level.
+            from slixmpp.stanza.stream_error import StreamError
+            error = StreamError()
+            error['condition'] = 'not-well-formed'
+            error['text'] = 'Server sent: %r' % data
+            self.send(error)
+            self.disconnect()
+
+    def is_connecting(self):
+        return self._current_connection_attempt is not None
 
     def is_connected(self):
         return self.transport is not None
@@ -399,7 +455,7 @@ class XMLStream(asyncio.BaseProtocol):
         closure of the TCP connection
         """
         log.info("connection_lost: %s", (exception,))
-        self.event("disconnected")
+        self.event("disconnected", self.disconnect_reason or exception and exception.strerror)
         if self.end_session_on_disconnect:
             self.event('session_end')
         # All these objects are associated with one TCP connection.  Since
@@ -408,42 +464,67 @@ class XMLStream(asyncio.BaseProtocol):
         self.transport = None
         self.socket = None
 
-    def disconnect(self, wait=2.0):
+    def cancel_connection_attempt(self):
+        """
+        Immediately cancel the current create_connection() Future.
+        This is useful when a client using slixmpp tries to connect
+        on flaky networks, where sometimes a connection just gets lost
+        and it needs to reconnect while the attempt is still ongoing.
+        """
+        if self._current_connection_attempt:
+            self._current_connection_attempt.cancel()
+            self._current_connection_attempt = None
+
+    def disconnect(self, wait: float = 2.0, reason: Optional[str] = None) -> None:
         """Close the XML stream and wait for an acknowldgement from the server for
         at most `wait` seconds.  After the given number of seconds has
-        passed without a response from the serveur, or when the server
-        successfuly responds with a closure of its own stream, abort() is
-        called. If wait is 0.0, this is almost equivalent to calling abort()
-        directly.
+        passed without a response from the server, or when the server
+        successfully responds with a closure of its own stream, abort() is
+        called. If wait is 0.0, this will call abort() directly without closing
+        the stream.
 
         Does nothing if we are not connected.
 
         :param wait: Time to wait for a response from the server.
 
         """
+        # Compat: docs/getting_started/sendlogout.rst has been promoting
+        # `disconnect(wait=True)` for ages. This doesn't mean anything to the
+        # schedule call below. It would fortunately be converted to `1` later
+        # down the call chain. Praise the implicit casts lord.
+        if wait == True:
+            wait = 2.0
+
+        self.disconnect_reason = reason
+        self.cancel_connection_attempt()
         if self.transport:
-            self.send_raw(self.stream_footer)
+            if wait > 0.0:
+                self.send_raw(self.stream_footer)
             self.schedule('Disconnect wait', wait,
                           self.abort, repeat=False)
+        else:
+            self.event("disconnected", reason)
 
     def abort(self):
         """
         Forcibly close the connection
         """
+        self.cancel_connection_attempt()
         if self.transport:
             self.transport.close()
             self.transport.abort()
             self.event("killed")
             self.disconnected.set_result(True)
             self.disconnected = asyncio.Future()
+            self.event("disconnected", self.disconnect_reason)
 
-    def reconnect(self, wait=2.0):
+    def reconnect(self, wait=2.0, reason="Reconnecting"):
         """Calls disconnect(), and once we are disconnected (after the timeout, or
         when the server acknowledgement is received), call connect()
         """
         log.debug("reconnecting...")
-        self.disconnect(wait)
-        self.add_event_handler('disconnected', self.connect, disposable=True)
+        self.add_event_handler('disconnected', lambda event: self.connect(), disposable=True)
+        self.disconnect(wait, reason)
 
     def configure_socket(self):
         """Set timeout and other options for self.socket.
@@ -469,14 +550,10 @@ class XMLStream(asyncio.BaseProtocol):
         """
         pass
 
-    def start_tls(self):
-        """Perform handshakes for TLS.
-
-        If the handshake is successful, the XML stream will need
-        to be restarted.
+    def get_ssl_context(self):
+        """
+        Get SSL context.
         """
-        self.event_when_connected = "tls_success"
-
         if self.ciphers is not None:
             self.ssl_context.set_ciphers(self.ciphers)
         if self.keyfile and self.certfile:
@@ -491,26 +568,44 @@ class XMLStream(asyncio.BaseProtocol):
             self.ssl_context.verify_mode = ssl.CERT_REQUIRED
             self.ssl_context.load_verify_locations(cafile=self.ca_certs)
 
-        ssl_connect_routine = self.loop.create_connection(lambda: self, ssl=self.ssl_context,
-                                                          sock=self.socket,
-                                                          server_hostname=self.default_domain)
-        @asyncio.coroutine
-        def ssl_coro():
-            try:
-                transp, prot = yield from ssl_connect_routine
-            except ssl.SSLError as e:
-                log.debug('SSL: Unable to connect', exc_info=True)
-                log.error('CERT: Invalid certificate trust chain.')
-                if not self.event_handled('ssl_invalid_chain'):
-                    self.disconnect()
-                else:
-                    self.event('ssl_invalid_chain', e)
-            else:
-                der_cert = transp.get_extra_info("socket").getpeercert(True)
-                pem_cert = ssl.DER_cert_to_PEM_cert(der_cert)
-                self.event('ssl_cert', pem_cert)
+        return self.ssl_context
 
-        asyncio.async(ssl_coro())
+    async def start_tls(self):
+        """Perform handshakes for TLS.
+
+        If the handshake is successful, the XML stream will need
+        to be restarted.
+        """
+        self.event_when_connected = "tls_success"
+        ssl_context = self.get_ssl_context()
+        try:
+            if hasattr(self.loop, 'start_tls'):
+                transp = await self.loop.start_tls(self.transport,
+                                                   self, ssl_context)
+            # Python < 3.7
+            else:
+                transp, _ = await self.loop.create_connection(
+                    lambda: self,
+                    ssl=self.ssl_context,
+                    sock=self.socket,
+                    server_hostname=self.default_domain
+                )
+        except ssl.SSLError as e:
+            log.debug('SSL: Unable to connect', exc_info=True)
+            log.error('CERT: Invalid certificate trust chain.')
+            if not self.event_handled('ssl_invalid_chain'):
+                self.disconnect()
+            else:
+                self.event('ssl_invalid_chain', e)
+            return False
+        der_cert = transp.get_extra_info("ssl_object").getpeercert(True)
+        pem_cert = ssl.DER_cert_to_PEM_cert(der_cert)
+        self.event('ssl_cert', pem_cert)
+        # If we use the builtin start_tls, the connection_made() protocol
+        # method is not called automatically
+        if hasattr(self.loop, 'start_tls'):
+            self.connection_made(transp)
+        return True
 
     def _start_keepalive(self, event):
         """Begin sending whitespace periodically to keep the connection alive.
@@ -623,8 +718,7 @@ class XMLStream(asyncio.BaseProtocol):
             idx += 1
         return False
 
-    @asyncio.coroutine
-    def get_dns_records(self, domain, port=None):
+    async def get_dns_records(self, domain, port=None):
         """Get the DNS records for a domain.
 
         :param domain: The domain in question.
@@ -636,7 +730,7 @@ class XMLStream(asyncio.BaseProtocol):
         resolver = default_resolver(loop=self.loop)
         self.configure_dns(resolver, domain=domain, port=port)
 
-        result = yield from resolve(domain, port,
+        result = await resolve(domain, port,
                                     service=self.dns_service,
                                     resolver=resolver,
                                     use_ipv6=self.use_ipv6,
@@ -644,8 +738,7 @@ class XMLStream(asyncio.BaseProtocol):
                                     loop=self.loop)
         return result
 
-    @asyncio.coroutine
-    def pick_dns_answer(self, domain, port=None):
+    async def pick_dns_answer(self, domain, port=None):
         """Pick a server and port from DNS answers.
 
         Gets DNS answers if none available.
@@ -655,7 +748,7 @@ class XMLStream(asyncio.BaseProtocol):
         :param port: If the results don't include a port, use this one.
         """
         if self.dns_answers is None:
-            dns_records = yield from self.get_dns_records(domain, port)
+            dns_records = await self.get_dns_records(domain, port)
             self.dns_answers = iter(dns_records)
 
         try:
@@ -719,17 +812,19 @@ class XMLStream(asyncio.BaseProtocol):
 
             # If the callback is a coroutine, schedule it instead of
             # running it directly
-            if asyncio.iscoroutinefunction(handler_callback):
-                @asyncio.coroutine
-                def handler_callback_routine(cb):
+            if iscoroutinefunction(handler_callback):
+                async def handler_callback_routine(cb):
                     try:
-                        yield from cb(data)
+                        await cb(data)
                     except Exception as e:
                         if old_exception:
                             old_exception(e)
                         else:
                             self.exception(e)
-                asyncio.async(handler_callback_routine(handler_callback))
+                asyncio.ensure_future(
+                    handler_callback_routine(handler_callback),
+                    loop=self.loop,
+                )
             else:
                 try:
                     handler_callback(data)
@@ -816,11 +911,93 @@ class XMLStream(asyncio.BaseProtocol):
         """
         return xml
 
+    async def _continue_slow_send(
+            self,
+            task: asyncio.Task,
+            already_used: Set[Callable[[ElementBase], Optional[StanzaBase]]]
+    ) -> None:
+        """
+        Used when an item in the send queue has taken too long to process.
+
+        This is away from the send queue and can take as much time as needed.
+        :param asyncio.Task task: the Task wrapping the coroutine
+        :param set already_used: Filters already used on this outgoing stanza
+        """
+        data = await task
+        for filter in self.__filters['out']:
+            if filter in already_used:
+                continue
+            if iscoroutinefunction(filter):
+                data = await task
+            else:
+                data = filter(data)
+            if data is None:
+                return
+
+        if isinstance(data, ElementBase):
+            for filter in self.__filters['out_sync']:
+                data = filter(data)
+                if data is None:
+                    return
+            str_data = tostring(data.xml, xmlns=self.default_ns,
+                                stream=self, top_level=True)
+            self.send_raw(str_data)
+        else:
+            self.send_raw(data)
+
+
+    async def run_filters(self):
+        """
+        Background loop that processes stanzas to send.
+        """
+        while True:
+            (data, use_filters) = await self.waiting_queue.get()
+            try:
+                if isinstance(data, ElementBase):
+                    if use_filters:
+                        already_run_filters = set()
+                        for filter in self.__filters['out']:
+                            already_run_filters.add(filter)
+                            if iscoroutinefunction(filter):
+                                task = asyncio.create_task(filter(data))
+                                completed, pending = await wait(
+                                    {task},
+                                    timeout=1,
+                                )
+                                if pending:
+                                    asyncio.ensure_future(
+                                        self._continue_slow_send(
+                                            task,
+                                            already_run_filters
+                                        )
+                                    )
+                                    raise Exception("Slow coro, rescheduling")
+                                data = task.result()
+                            else:
+                                data = filter(data)
+                            if data is None:
+                                raise ContinueQueue('Empty stanza')
+
+                if isinstance(data, ElementBase):
+                    if use_filters:
+                        for filter in self.__filters['out_sync']:
+                            data = filter(data)
+                            if data is None:
+                                raise ContinueQueue('Empty stanza')
+                    str_data = tostring(data.xml, xmlns=self.default_ns,
+                                        stream=self, top_level=True)
+                    self.send_raw(str_data)
+                else:
+                    self.send_raw(data)
+            except ContinueQueue as exc:
+                log.debug('Stanza in send queue not sent: %s', exc)
+            except Exception:
+                log.error('Exception raised in send queue:', exc_info=True)
+            self.waiting_queue.task_done()
+
     def send(self, data, use_filters=True):
         """A wrapper for :meth:`send_raw()` for sending stanza objects.
 
-        May optionally block until an expected response is received.
-
         :param data: The :class:`~slixmpp.xmlstream.stanzabase.ElementBase`
                      stanza to send on the stream.
         :param bool use_filters: Indicates if outgoing filters should be
@@ -828,25 +1005,7 @@ class XMLStream(asyncio.BaseProtocol):
                                  filters is useful when resending stanzas.
                                  Defaults to ``True``.
         """
-        if isinstance(data, ElementBase):
-            if use_filters:
-                for filter in self.__filters['out']:
-                    data = filter(data)
-                    if data is None:
-                        return
-
-        if isinstance(data, ElementBase):
-            if use_filters:
-                for filter in self.__filters['out_sync']:
-                    data = filter(data)
-                    if data is None:
-                        return
-            str_data = tostring(data.xml, xmlns=self.default_ns,
-                                          stream=self,
-                                          top_level=True)
-            self.send_raw(str_data)
-        else:
-            self.send_raw(data)
+        self.waiting_queue.put_nowait((data, use_filters))
 
     def send_xml(self, data):
         """Send an XML object on the stream
@@ -861,7 +1020,7 @@ class XMLStream(asyncio.BaseProtocol):
 
         :param string data: Any bytes or utf-8 string value.
         """
-        log.debug("[36;1mSEND[0m: %s", highlight(data))
+        log.debug("SEND: %s", data)
         if not self.transport:
             raise NotConnectedError()
         if isinstance(data, str):
@@ -914,7 +1073,7 @@ class XMLStream(asyncio.BaseProtocol):
         if stanza is None:
             return
 
-        log.debug("[33;1mRECV[0m: %s", highlight(stanza))
+        log.debug("RECV: %s", stanza)
 
         # Match the stanza against registered handlers. Handlers marked
         # to run "in stream" will be executed immediately; the rest will
@@ -944,4 +1103,3 @@ class XMLStream(asyncio.BaseProtocol):
         :param exception: An unhandled exception object.
         """
         pass
-
